import cv2
import numpy as np
# Fix horrible naming of OpenCV functions
from cv2.fisheye import (
    estimateNewCameraMatrixForUndistortRectify as estimateCameraMatrix,
    initUndistortRectifyMap,
    undistortPoints,
    distortPoints,
)
from numpy.linalg import inv
from shapely.geometry import LineString, Point


def wpca(X, w):
    """ Weighted PCA """
    U, E, _ = np.linalg.svd(X.T @ np.diag(w) @ X)
    return U, E


def fit_line(x):
    n = x.shape[0]
    pts = np.vstack([x.T, np.ones((1, n))])
    pts = pts / np.linalg.norm(pts, axis=0, keepdims=True)
    v, w = wpca(pts.T, np.ones(n))
    p = v[:, np.argmin(w)]
    return p / p[2]  # Line params


def line_segment(l, x0, x1):
    left = np.array([-1, 0, x0])
    right = np.array([-1, 0, x1])

    hl = np.cross(l, left)
    hr = np.cross(l, right)

    x1, y1 = hl[:2] / hl[2]
    x2, y2 = hr[:2] / hr[2]

    return x1, y1, x2, y2


class Camera:
    def __init__(self, image_size, rectified_size, K, D, RT=None):
        self.horizon = None
        self.RT_inv = None
        self.image_size = image_size
        self.rectified_size = rectified_size
        self.K = K
        self.D = D
        self.RT = RT or np.eye(3, 4)

        # Rectification parameters
        self.K_new = estimateCameraMatrix(
            self.K, self.D, tuple(self.image_size), np.eye(3), new_size=tuple(self.rectified_size), fov_scale=1.2
        )
        self.maps = initUndistortRectifyMap(
            self.K, self.D, np.eye(3), self.K_new, tuple(self.rectified_size), cv2.CV_32F
        )

        # view_direction = d.get("view_direction", "x")
        # R = np.eye(4)
        # R[:3,:3] = estimate_R(d["K"], d["horizon"], view_direction)
        # T = translation_matrix(d.get("location", [0,0,1]))
        # self.RT = inv(self.T @ self.R)

    def project_points(self, X, near: float = 0, to_rectified: bool = True):
        """
        X : (N,3) matrix
        """
        n = X.shape[0]
        X = np.vstack([X.T, np.ones((1, n))])
        K = self.K_new if to_rectified else self.K
        x = K @ self.RT @ X
        d = x[2]
        valid = d > near
        x = x[:2, valid] / d[valid]
        return x.T, d[valid]

    def rectify_image(self, image):
        map1, map2 = self.maps
        img = cv2.GaussianBlur(image, (3, 3), 0.5)
        return cv2.remap(img, map1, map2, cv2.INTER_NEAREST, borderMode=cv2.BORDER_CONSTANT)

    def rectify_points(self, x):
        y = undistortPoints(x.reshape(1, -1, 2), self.K, self.D, P=self.K_new)
        return y[0]

    def unrectify_points(self, x):
        """
        x : (n,2) in K_new
        """
        n = x.shape[0]
        x = np.vstack([x.T, np.ones((1, n))]).astype(np.float32)
        x_norm = inv(self.K_new) @ x
        y = distortPoints(x_norm.reshape(1, -1, 2), self.K, self.D)
        return y[0]

    @staticmethod
    def from_dict(d: dict) -> "Camera":
        image_size = d["image_size"]
        rectified_size = d["rectified_size"]
        # Measured intrinsic matrix
        K = np.array(d["K"], "f")
        # Measured distortion coefficients
        D = np.array(d["D"], "f")
        cam = Camera(image_size, rectified_size, K, D)

        w, h = image_size
        rw, rh = rectified_size

        # Estimate rotation matrix
        # Get points on horizon or simply one point in image center
        h_points = np.atleast_2d(d.get("horizon_points", [(w / 2, h / 2)])).astype(np.float32)
        # Get undistorted coords of the points
        h_points = cam.rectify_points(h_points)
        if h_points.shape[0] == 1:  # Single point case - add second point on the same line
            x, y = h_points[0]
            h_points = np.vstack([h_points, (x + 100, y)])
        # Line segment in undistorted image defining the horizon
        x1, y1, x2, y2 = line_segment(fit_line(h_points), 0, rw)
        # Represent horizon as LineString - convenient for geometric processing
        cam.horizon = LineString([(x1, y1), (x2, y2)])
        # Get location of first point on the horizon - the closest point on the line
        x1, y1 = cam.horizon.interpolate(cam.horizon.project(Point(h_points[0]))).coords[0]
        # Estimate the rotation matrix
        R = np.eye(4)
        R[:3, :3] = estimate_R(cam.K_new, (x1, y1, x2, y2), d.get("view_direction", "x"))
        # add translation
        T = translation_matrix(d.get("location", [0, 0, 1]))
        cam.RT = np.linalg.inv(T @ R)[:3]
        cam.RT_inv = (T @ R)[:3]

        return cam


def estimate_R(K, h, view_direction="x"):
    """
    Estimate rotation matrix from horizon observation
    """
    assert view_direction in {"x", "-x"}
    x1, y1, x2, y2 = h
    # Compose matrix with homogeneous points
    h = np.array(
        [
            [x1, y1, 1],  # Point in X direction
            [x2, y2, 1]  # Point elsewhere on the horizon
        ]
    ).T

    # Directions to the points
    H = (np.linalg.inv(K) @ h).T

    # Calc direction of the individual axes

    direct_sign = +1 if view_direction == "x" else -1
    direct = direct_sign * H[0]

    up = np.cross(H[1], H[0])
    up = -np.sign(up[1]) * up

    right = np.cross(up, direct)

    R = np.array([direct, right, up])
    n = np.linalg.norm(R, axis=1, keepdims=True)

    return R / n


def translation_matrix(t):
    T = np.eye(4)
    T[:3, -1] = t
    return T



from __future__ import annotations

import csv
import logging
import os
import statistics
import time
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Callable, Optional

import numpy as np
import yaml
from fcw.core.geometry import Camera

from era_5g_client.client_base import NetAppClientBase

logger = logging.getLogger(__name__)

image_storage: Dict[int, np.ndarray] = dict()

DEBUG_PRINT_WARNING = True  # Prints score.
DEBUG_PRINT_DELAY = True  # Prints the delay between capturing image and receiving the results.

# URL of the FCW service
NETAPP_ADDRESS = str(os.getenv("NETAPP_ADDRESS", "http://localhost:5896"))


class ResultsReader:
    """Default class for processing FCW results."""

    def __init__(self, out_csv_dir: str = None, out_prefix: str = None) -> None:
        """Constructor.

        Args:
            out_csv_dir (str, optional): Dir for CSV timestamps stats.
            out_prefix (str, optional): Filename prefix for CSV timestamps stats.
        """

        self.delays = []
        self.delays_recv = []
        self.delays_send = []
        self.timestamps = [
            ["start_timestamp_ns",
             "recv_timestamp_ns",
             "send_timestamp_ns",
             "end_timestamp_ns"]
        ]
        self.out_csv_dir = out_csv_dir
        self.out_prefix = out_prefix

    def stats(self, send_frames_count: int) -> None:
        """Print timestamps stats and can write them to CSV file.

        Args:
            send_frames_count (int): Number of send frames.
        """

        logger.info(f"-----")
        if len(self.delays) < 1 or len(self.delays_recv) < 1 or len(self.delays_send) < 1:
            logger.warning(f"No results data received")
        else:
            logger.info(f"Send frames: {send_frames_count}, dropped frames: {send_frames_count - len(self.delays)}")
            logger.info(
                f"Delay median: {statistics.median(self.delays) * 1.0e-9:.3f}s "
                f"mean: {statistics.mean(self.delays) * 1.0e-9:.3f}s "
                f"min: {min(self.delays) * 1.0e-9:.3f}s "
                f"max: {max(self.delays) * 1.0e-9:.3f}s"
            )
            # logger.info(
            #     f"Delay service recv median: {statistics.median(self.delays_recv) * 1.0e-9:.3f}s "
            #     f"mean: {statistics.mean(self.delays_recv) * 1.0e-9:.3f}s "
            #     f"min: {min(self.delays_recv) * 1.0e-9:.3f}s "
            #     f"max: {max(self.delays_recv) * 1.0e-9:.3f}s"
            # )
            # logger.info(
            #     f"Delay service send median: {statistics.median(self.delays_send) * 1.0e-9:.3f}s "
            #     f"mean: {statistics.mean(self.delays_send) * 1.0e-9:.3f}s "
            #     f"min: {min(self.delays_send) * 1.0e-9:.3f}s "
            #     f"max: {max(self.delays_send) * 1.0e-9:.3f}s"
            # )
            if self.out_csv_dir is not None:
                out_csv_filename = f'{self.out_prefix}'
                out_csv_filepath = os.path.join(self.out_csv_dir, out_csv_filename + ".csv")
                with open(out_csv_filepath, "w", newline='') as csv_file:
                    csv_writer = csv.writer(csv_file)
                    csv_writer.writerows(self.timestamps)

    def get_results(self, results: Dict[str, Any]) -> None:
        """Callback which process the results from the FCW service.

        Args:
            results (Dict[str, Any]): The results in JSON format.
        """

        results_timestamp = time.perf_counter_ns()

        # Process detections
        if "detections" in results:
            if DEBUG_PRINT_WARNING:
                for tracked_id, detection in results["detections"].items():
                    score = float(detection["dangerous_distance"])
                    if score > 0:
                        logger.info(f"Dangerous distance {score:.2f}m to the object with id {tracked_id}")

        # Process timestamps
        if "timestamp" in results:
            timestamp = results["timestamp"]
            recv_timestamp = results["recv_timestamp"]
            send_timestamp = results["send_timestamp"]

            if DEBUG_PRINT_DELAY:
                logger.info(
                    f"Result number {len(self.timestamps)}"
                    f", delay: {(results_timestamp - timestamp) * 1.0e-9:.3f}s"
                    #f", recv frame delay: {(recv_timestamp - timestamp) * 1.0e-9:.3f}s"
                )
                self.delays.append((results_timestamp - timestamp))
                self.delays_recv.append((recv_timestamp - timestamp))
                self.delays_send.append((send_timestamp - timestamp))

            self.timestamps.append(
                [
                    timestamp,
                    recv_timestamp,
                    send_timestamp,
                    results_timestamp
                ]
            )


class StreamType(Enum):
    JPEG = 1
    H264 = 2


class CollisionWarningClient:
    """Wrapper class for FCW client."""

    def __init__(
        self,
        config: Path,
        camera_config: Path,
        netapp_address: str = NETAPP_ADDRESS,
        fps: float = 30,
        results_callback: Optional[Callable] = None,
        stream_type: Optional[StreamType] = StreamType.H264,
        out_csv_dir: Optional[str] = None,
        out_prefix: Optional[str] = "fcw_test_",
    ) -> None:
        """Constructor.

        Args:
            config (Path): Path to FCW configuration file.
            camera_config (Path): Path to camera configuration file.
            netapp_address (str, optional): The URI and port of the FCW service interface.
                Default taken from environment variables NETAPP_ADDRESS and NETAPP_PORT.
            fps (float, optional): Video FPS. Default to 30.
            results_callback (Callable, optional): Callback for receiving results.
                Default to ResultsReader.get_results
            stream_type (StreamType, optional): Stream type JPEG or H264. Default to H264.
            out_csv_dir (str, optional): Dir for CSV timestamps stats. Default to None.
            out_prefix (str, optional): Filename prefix for CSV timestamps stats. Default to "fcw_test_".
        """

        logger.info("Loading configuration file {cfg}".format(cfg=config))
        self.config_dict = yaml.safe_load(config.open())
        logger.info("Loading camera configuration {cfg}".format(cfg=camera_config))
        self.camera_config_dict = yaml.safe_load(camera_config.open())
        logger.info("Initializing camera calibration")
        self.camera = Camera.from_dict(self.camera_config_dict)

        width, height = self.camera.rectified_size
        self.fps = fps
        # Check bad loaded FPS
        if self.fps > 60:
            logger.warning(f"FPS {self.fps} is strangely high, newly set to 30")
            self.fps = 30
        self.results_callback = results_callback
        if self.results_callback is None:
            self.results_viewer = ResultsReader(
                out_csv_dir=out_csv_dir, out_prefix=out_prefix
            )
            self.results_callback = self.results_viewer.get_results
        self.stream_type = stream_type
        self.frame_id = 0

        # Create FCW client
        self.client = NetAppClientBase(self.results_callback)
        logger.info(
            f"Register with netapp_address: {netapp_address}"
        )
        # Register client
        try:
            if self.stream_type is StreamType.H264:
                self.client.register(
                    netapp_address,
                    args={"h264": True, "config": self.config_dict, "camera_config": self.camera_config_dict,
                          "fps": self.fps,
                          "width": width, "height": height}
                )
            elif self.stream_type is StreamType.JPEG:
                self.client.register(
                    netapp_address,
                    args={"config": self.config_dict, "camera_config": self.camera_config_dict, "fps": self.fps}
                )
            else:
                raise Exception("Unknown stream type")
        except Exception as e:
            self.client.disconnect()
            raise e

    def send_image(self, frame: np.ndarray, timestamp: Optional[int] = None) -> None:
        """Send image to FCW service including rectification.

        Args:
            frame (np.ndarray): Image in numpy array format ("bgr24").
            timestamp (int, optional): Timestamp for frame and results synchronization.
                Default to current time (time.perf_counter_ns())
        """

        if self.client is not None:
            self.frame_id += 1
            frame_undistorted = self.camera.rectify_image(frame)
            if not timestamp:
                timestamp = time.perf_counter_ns()
            self.client.send_image_ws(frame_undistorted, timestamp)

    def stop(self) -> None:
        """Print stats and disconnect from FCW service."""
        logger.info("Collision warning client stopping")

        if hasattr(self, "results_viewer") and self.results_viewer is not None:
            self.results_viewer.stats(self.frame_id)
        if self.client is not None:
            self.client.disconnect()



from __future__ import annotations
import signal
import time
import traceback
from argparse import ArgumentParser
from pathlib import Path
import cv2
import sys
import logging
import os

# FCW and 5G-ERA stuff
from fcw.client_python.client_common import CollisionWarningClient
from fcw.client_python.client_common import StreamType
from era_5g_client.exceptions import FailedToConnect
from era_5g_interface.utils.rate_timer import RateTimer

# Set logging
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("FCW client python")

# Testing video file
TEST_VIDEO_FILE = str("../../videos/video3.mp4")
# TEST_VIDEO_FILE = str("rtsp://127.0.0.1:8554/webcam.h264")
# TEST_VIDEO_FILE = str("../../videos/bringauto_2023-03-20.mp4")

# Testing configuration of the algorithm
CONFIG_FILE = Path("../../config/config.yaml")

# Testing camera settings - specific for the particular input
CAMERA_CONFIG_FILE = Path("../../videos/video3.yaml")
# CAMERA_CONFIG_FILE = Path("../../videos/bringauto.yaml")

# topped flag for SIGTERM handler (stopping video frames sending)
stopped = False
collision_warning_client = None


def signal_handler(sig: int, frame) -> None:
    """Signal handler for SIGTERM and SIGINT."""
    logger.info(f"Terminating ({signal.Signals(sig).name})...")
    global stopped
    stopped = True
    # collision_warning_client.stop()


signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGINT, signal_handler)


def main() -> None:
    """Example of FCW client."""

    # Parse arguments
    parser = ArgumentParser()
    parser.add_argument(
        "-s", "--stream_type", type=int, help="StreamType: 1 = JPEG, 2 = H264",
        default=StreamType.H264
    )
    parser.add_argument("-c", "--config", type=Path, help="Collision warning config", default=CONFIG_FILE)
    parser.add_argument("--camera", type=Path, help="Camera settings", default=CAMERA_CONFIG_FILE)
    parser.add_argument("-o", "--out_csv_dir", type=str, help="Output CSV dir", default=None)
    parser.add_argument(
        "-p", "--out_prefix", type=str, help="Prefix of output csv file with measurements", default=None
    )
    parser.add_argument("-t", "--play_time", type=int, help="Video play time in seconds", default=5000)
    parser.add_argument("--fps", type=int, help="Video FPS", default=None)
    parser.add_argument("source_video", type=str, help="Video stream (file or url)", nargs='?', default=TEST_VIDEO_FILE)
    args = parser.parse_args()

    global collision_warning_client

    try:
        # Create a video capture to pass images to the NetApp
        logger.info(f"Opening video capture {args.source_video}")
        cap = cv2.VideoCapture(args.source_video)
        if not cap.isOpened():
            raise Exception("Cannot open video capture")

        # Check bad loaded FPS
        if not args.fps:
            fps = cap.get(cv2.CAP_PROP_FPS)
            if fps > 60:
                logger.warning(f"FPS {fps} is strangely high, newly set to 30")
                fps = 30
        else:
            fps = args.fps

        # Get width, height and number of frames
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        # Number of frames can be wrong
        frames = cap.get(cv2.CAP_PROP_FRAME_COUNT)

        logger.info(
            f"Video {width}x{height}, {fps} FPS, "
            f"{frames} frames, "
            f"{frames / fps} seconds"
        )
        logger.info(f"Stream type: {StreamType(args.stream_type)}")

        # Create collision warning client with given parameters,
        # width and height for client registration is loaded from camera config
        collision_warning_client = CollisionWarningClient(
            config=args.config, camera_config=args.camera, fps=fps, stream_type=StreamType(args.stream_type),
            out_csv_dir=args.out_csv_dir, out_prefix=args.out_prefix
        )

        # Rate timer for control the speed of a loop (fps)
        rate_timer = RateTimer(rate=fps, time_function=time.perf_counter, iteration_miss_warning=True)
        # Start time
        start_time = time.perf_counter_ns()
        # Check elapsed time or stopped flag
        # Play time in ns
        play_time_ns = args.play_time * 1.0e+9
        # Main processing loop
        while time.perf_counter_ns() - start_time < play_time_ns and not stopped:
            # Read single frame from a stream
            ret, frame = cap.read()
            if not ret:
                break
            # Send to FCW service for detection/processing
            collision_warning_client.send_image(frame)
            # Sleep until next frame should be sent (with given fps)
            rate_timer.sleep()
        end_time = time.perf_counter_ns()
        logger.info(f"Total streaming time: {(end_time - start_time) * 1.0e-9:.3f}s")
        cap.release()

    except FailedToConnect as ex:
        logger.error(f"Failed to connect to server: {ex}")
        sys.exit(1)
    except KeyboardInterrupt:
        logger.info("Terminating ...")
    except Exception as ex:
        traceback.print_exc()
        logger.error(f"Exception: {repr(ex)}")
        sys.exit(1)
    finally:
        if collision_warning_client is not None:
            collision_warning_client.stop()


if __name__ == "__main__":
    main()



from __future__ import annotations

import traceback
from argparse import ArgumentParser
from pathlib import Path
import cv2
import sys
import logging
from typing import Any, Dict

# FCW and 5G-ERA stuff
from fcw.client_python.client_common import CollisionWarningClient
from era_5g_client.exceptions import FailedToConnect
from era_5g_interface.utils.rate_timer import RateTimer

# Set logging
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("FCW client python")

# Testing video file
TEST_VIDEO_FILE = str("../../videos/video3.mp4")
# TEST_VIDEO_FILE = str("rtsp://127.0.0.1:8554/webcam.h264")

# Testing configuration of the algorithm
CONFIG_FILE = Path("../../config/config.yaml")

# Testing camera settings - specific for the particular input
CAMERA_CONFIG_FILE = Path("../../videos/video3.yaml")


def results_callback(results: Dict[str, Any]) -> None:
    """Callback which process the results from the FCW service.

        Args:
            results (Dict[str, Any]): The results in JSON format.
        """

    # Process detections
    if "detections" in results:
        for tracked_id, detection in results["detections"].items():
            score = float(detection["dangerous_distance"])
            if score > 0:
                logger.info(f"Dangerous distance {score:.2f}m to the object with id {tracked_id}")


def main() -> None:
    """Example of simple FCW client."""

    # Parse arguments
    parser = ArgumentParser()
    parser.add_argument("-c", "--config", type=Path, help="Collision warning config", default=CONFIG_FILE)
    parser.add_argument("--camera", type=Path, help="Camera settings", default=CAMERA_CONFIG_FILE)
    parser.add_argument("source_video", type=str, help="Video stream (file or url)", nargs='?', default=TEST_VIDEO_FILE)
    args = parser.parse_args()

    collision_warning_client = None

    try:
        # Create a video capture to pass images to the NetApp
        logger.info(f"Opening video capture {args.source_video}")
        cap = cv2.VideoCapture(args.source_video)
        if not cap.isOpened():
            raise Exception("Cannot open video capture")

        # FPS can be wrong, it should be checked
        fps = cap.get(cv2.CAP_PROP_FPS)

        # Create collision warning client with given parameters,
        # width and height for client registration is loaded from camera config
        collision_warning_client = CollisionWarningClient(
            config=args.config,
            camera_config=args.camera, fps=fps, results_callback=results_callback
        )

        # Rate timer for control the speed of a loop (fps)
        rate_timer = RateTimer(rate=fps)

        # Main processing loop
        while True:
            # Read single frame from a stream
            ret, frame = cap.read()
            if not ret:
                break
            # Send to FCW service for detection/processing
            collision_warning_client.send_image(frame)
            # Sleep until next frame should be sent (with given fps)
            rate_timer.sleep()
        cap.release()

    except FailedToConnect as ex:
        logger.error(f"Failed to connect to server: {ex}")
        sys.exit(1)
    except KeyboardInterrupt:
        logger.info("Terminating...")
    except Exception as ex:
        traceback.print_exc()
        logger.error(f"Exception: {repr(ex)}")
        sys.exit(1)
    finally:
        if collision_warning_client is not None:
            collision_warning_client.stop()


if __name__ == "__main__":
    main()



import base64
import argparse
import binascii
import os
import numpy as np
import time
import socketio
from dataclasses import dataclass
from queue import Queue
from typing import Dict
import threading
import cv2
from flask import Flask
import logging
import sys

from fcw.service.collision_worker import CollisionWorker

from era_5g_interface.task_handler import TaskHandler
from era_5g_interface.task_handler_internal_q import TaskHandlerInternalQ
from era_5g_interface.h264_decoder import H264Decoder
import era_5g_interface.interface_helpers
from era_5g_interface.interface_helpers import HeartBeatSender
from era_5g_interface.dataclasses.control_command import ControlCommand, ControlCmdType

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("FCW interface")

# port of the netapp's server
NETAPP_PORT = os.getenv("NETAPP_PORT", 5896)
# input queue size
NETAPP_INPUT_QUEUE = int(os.getenv("NETAPP_INPUT_QUEUE", 1))

# the max_http_buffer_size parameter defines the max size of the message to be passed
sio = socketio.Server(async_mode='threading', async_handlers=False, max_http_buffer_size=5 * (1024 ** 2))
app = Flask(__name__)
app.wsgi_app = socketio.WSGIApp(sio, app.wsgi_app)


@dataclass
class TaskAndWorker:
    task: TaskHandler
    worker: CollisionWorker


# list of registered tasks
tasks: Dict[str, TaskAndWorker] = dict()

heart_beat_sender = HeartBeatSender()


def heart_beat_timer():
    latencies = []
    for task_and_worker in tasks.values():
        latencies.extend(task_and_worker.worker.latency_measurements.get_latencies())
    avg_latency = 0
    if len(latencies) > 0:
        avg_latency = float(np.mean(np.array(latencies)))

    queue_size = 1
    queue_occupancy = 1

    heart_beat_sender.send_middleware_heart_beat(
        avg_latency=avg_latency, queue_size=queue_size, queue_occupancy=queue_occupancy, current_robot_count=len(tasks)
    )
    threading.Timer(era_5g_interface.interface_helpers.MIDDLEWARE_REPORT_INTERVAL, heart_beat_timer).start()


heart_beat_timer()


def get_sid_of_namespace(eio_sid, namespace):
    return sio.manager.sid_from_eio_sid(eio_sid, namespace)


def get_results_sid(eio_sid):
    return sio.manager.sid_from_eio_sid(eio_sid, "/results")


@sio.on('connect', namespace='/data')
def connect_data(sid, environ):
    """Creates a websocket connection to the client for passing the data.

    Raises:
        ConnectionRefusedError: Raised when attempt for connection were made
            without registering first.
    """

    logger.info(f"Connected data. Session id: {sio.manager.eio_sid_from_sid(sid, '/data')}, namespace_id: {sid}")
    sio.send("You are connected", namespace='/data', to=sid)


@sio.on('connect', namespace='/control')
def connect_control(sid, environ):
    """_summary_
    Creates a websocket connection to the client for passing control commands.

    Raises:
        ConnectionRefusedError: Raised when attempt for connection were made
            without registering first.
    """

    logger.info(f"Connected control. Session id: {sio.manager.eio_sid_from_sid(sid, '/control')}, namespace_id: {sid}")
    sio.send("You are connected", namespace='/control', to=sid)


@sio.on('connect', namespace='/results')
def connect_results(sid, environ):
    """Creates a websocket connection to the client for passing the results.

    Raises:
        ConnectionRefusedError: Raised when attempt for connection were made
            without registering first.
    """

    logger.info(f"Connected results. Session id: {sio.manager.eio_sid_from_sid(sid, '/results')}, namespace_id: {sid}")
    sio.send("You are connected", namespace='/results', to=sid)


@sio.on('image', namespace='/data')
def image_callback_websocket(sid, data: dict):
    """Allows to receive jpg or h264 encoded image using the websocket transport

    Args:
        sid ():
        data (dict): A base64 encoded image frame and (optionally) related timestamp in format:
            {'frame': 'base64data', 'timestamp': 'int'}

    Raises:
        ConnectionRefusedError: Raised when attempt for connection were made
            without registering first or frame was not passed in correct format.
    """

    if 'timestamp' in data:
        timestamp = data['timestamp']
    else:
        logger.info("Timestamp not set, setting default value")
        timestamp = 0

    eio_sid = sio.manager.eio_sid_from_sid(sid, "/data")

    if eio_sid not in tasks:
        logger.error(f"Non-registered client tried to send data")
        sio.emit(
            "image_error",
            {"timestamp": timestamp,
             "error": "Not connected"},
            namespace='/data',
            to=sid
        )
        return

    if 'frame' not in data:
        logger.error(f"Data does not contain frame.")
        sio.emit(
            "image_error",
            {"timestamp": timestamp,
             "error": "Data does not contain frame."},
            namespace='/data',
            to=sid
        )
        return

    task = tasks[eio_sid].task

    last_timestamp = task.last_timestamp
    if timestamp - last_timestamp < 0:
        logger.error(
            f"Received frame with older timestamp: {timestamp}, "
            f"last_timestamp: {last_timestamp}, diff: {timestamp - last_timestamp}"
        )
        sio.emit(
            "image_error",
            {"timestamp": timestamp,
             "error": f"Received frame with older timestamp: {timestamp}, "
                      f"last_timestamp: {last_timestamp}, diff: {timestamp - last_timestamp}"},
            namespace='/data',
            to=sid
        )
        return
    tasks[eio_sid].task.last_timestamp = timestamp

    try:
        if task.decoder:
            image = task.decoder.decode_packet_data(data["frame"])
        else:
            frame = base64.b64decode(data["frame"])
            image = cv2.imdecode(np.frombuffer(frame, dtype=np.uint8), cv2.IMREAD_COLOR)
    except (ValueError, binascii.Error, Exception) as error:
        logger.error(f"Failed to decode frame data: {error}")
        sio.emit(
            "image_error",
            {"timestamp": timestamp,
             "error": f"Failed to decode frame data: {error}"},
            namespace='/data',
            to=sid
        )
        return

    task.store_image(
        {"sid": eio_sid,
         "timestamp": timestamp,
         "recv_timestamp": time.perf_counter_ns(),
         "websocket_id": get_results_sid(sio.manager.eio_sid_from_sid(sid, "/data"))},
        image
    )


@sio.on('json', namespace='/data')
def json_callback_websocket(sid, data):
    """
    Allows to receive general json data using the websocket transport

    Args:
        data (dict): NetApp-specific json data

    Raises:
        ConnectionRefusedError: Raised when attempt for connection were made
            without registering first.
    """
    print(data)

    logger.info(f"Client with task id: {sio.manager.eio_sid_from_sid(sid, '/data')} sent data {data}")


@sio.on('command', namespace='/control')
def command_callback_websocket(sid, data: Dict):
    command = ControlCommand(**data)
    logger.info(f"Control command {command} processing: session id: {sid}")
    if command and command.cmd_type == ControlCmdType.SET_STATE:
        args = command.data
        h264 = False
        config = {}
        camera_config = {}
        fps = 30
        width = 0
        height = 0
        if args:
            h264 = args.get("h264", False)
            config = args.get("config", {})
            camera_config = args.get("camera_config", {})
            fps = args.get("fps", 30)
            width = args.get("width", 0)
            height = args.get("height", 0)
            logger.info(f"H264: {h264}")
            logger.info(f"Config: {config}")
            logger.info(f"Camera config: {camera_config}")
            logger.info(f"Video {width}x{height}, {fps} FPS")

        # queue with received images
        image_queue = Queue(NETAPP_INPUT_QUEUE)
        eio_sid = sio.manager.eio_sid_from_sid(sid, '/control')

        if h264:
            task = TaskHandlerInternalQ(eio_sid, image_queue, decoder=H264Decoder(fps, width, height))
        else:
            task = TaskHandlerInternalQ(eio_sid, image_queue)

        # Create worker
        worker = CollisionWorker(
            image_queue, sio,
            config, camera_config, fps,
            name=f"Detector {eio_sid}"
        )

        tasks[eio_sid] = TaskAndWorker(task, worker)
        tasks[eio_sid].worker.start()
        t0 = time.perf_counter_ns()
        while True:
            if tasks[eio_sid].worker.is_alive():
                break
            if time.perf_counter_ns() > t0 + 5 * 1.0e+9:
                logger.error(f"Timed out to start worker. Session id: {eio_sid}, namespace_id: {sid}")
                raise ConnectionRefusedError('Timed out to start worker.')

        logger.info(f"Task handler and worker created and started: {eio_sid}")

    logger.info(f"Control command {command} applied: session id: {sid}")


@sio.on('disconnect', namespace='/data')
def disconnect_data(sid):
    eio_sid = sio.manager.eio_sid_from_sid(sid, "/data")
    task_and_worker = tasks.get(eio_sid)
    if task_and_worker:
        task_and_worker.worker.stop()
        task_and_worker.worker.join()
        del tasks[eio_sid]
        del task_and_worker
        logger.info(f"Task handler and worker deleted: {eio_sid}")

    logger.info(f"Client disconnected from /data namespace: session id: {sid}")


@sio.on('disconnect', namespace='/control')
def disconnect_control(sid):
    logger.info(f"Client disconnected from /control namespace: session id: {sid}")


@sio.on('disconnect', namespace='/results')
def disconnect_results(sid):
    logger.info(f"Client disconnected from /results namespace: session id: {sid}")


def main(args=None):
    parser = argparse.ArgumentParser(description='Standalone variant of Forward Collision Warning NetApp')
    args = parser.parse_args()

    logger.info(f"The size of the queue set to: {NETAPP_INPUT_QUEUE}")

    # runs the flask server
    # allow_unsafe_werkzeug needs to be true to run inside the docker
    # TODO: use better webserver
    app.run(port=NETAPP_PORT, host='0.0.0.0')


if __name__ == '__main__':
    main()



from dataclasses import dataclass
from typing import Iterable

import numpy as np
from shapely.geometry import Polygon


@dataclass(eq=False, frozen=True, order=False)
class ObjectObservation:
    geometry: Polygon
    score: float
    label: int

    def bounds(self):
        return self.geometry.bounds

    def numpy(self):
        return np.atleast_2d(self.bounds() + (self.score,))

    def is_in_frame(self, shape, margin=5):
        h, w = shape
        x1, y1, x2, y2 = self.bounds()
        return (x1 > margin) and (x2 < (w - margin)) and (y1 > margin) and (y2 < (h - margin))


def detections_to_numpy(detections: Iterable[ObjectObservation]) -> np.ndarray:
    detections = [d.numpy() for d in detections]
    detections.append(np.empty((0, 5)))  # vstack does not accept empty list - we add empty array, so it does not fail
    return np.vstack(detections)



"""
Wrapper for YOLO v5 detector from Ultralytics

COGNITECHNA spol. s r.o.
Roman Juranek <r.juranek@cognitechna.cz>

---

List of available classes can be found here:
https://github.com/ultralytics/yolov5/blob/master/data/coco128.yaml
"""

from typing import Iterable

import cv2
import numpy as np
import torch
from shapely.geometry import box
import gc
import logging
logger = logging.getLogger(__name__)

from fcw.core.detection import ObjectObservation


class YOLODetector:
    default_classes = [
        "person", "bicycle", "car", "motorcycle", "bus", "truck"
    ]

    def __init__(
        self,
        model: str = "yolov5l6",
        classes: Iterable[str] = None,
        max_size: int = 640,
        min_score: float = 0.3,
        filter_in_frame: bool = True,
        min_area: float = None
    ):
        self.model = torch.hub.load("ultralytics/yolov5", model, pretrained=True)
        self.model.agnostic = False
        self.model.iou = 0.7
        classes = classes or YOLODetector.default_classes
        if classes is not None:
            # Init detected classes 
            # Inverted name index: name -> class_id
            name_idx = dict(((name, class_id) for class_id, name in self.model.names.items()))
            # List class_id specified by names in classes passed as parameter, ignoring unknown classes
            self.model.classes = [name_idx[nm] for nm in classes if
                                  nm in name_idx] or None  # ... or None - in case og empty list leave None value not empty list
        self.model.conf = min_score
        self.max_size = max_size
        self.filter_in_frame = filter_in_frame
        self.min_area = min_area

    def __del__(self):
        self.memory_stats()
        logger.info(f"Free torch cuda memory ...")
        self.model.cpu()
        del self.model
        gc.collect()
        torch.cuda.empty_cache()
        self.memory_stats()

    @staticmethod
    def memory_stats():
        logger.info(f"torch.cuda.memory_allocated(): {torch.cuda.memory_allocated() / 1024 ** 2}")
        logger.info(f"torch.cuda.memory_cached(): {torch.cuda.memory_reserved() / 1024 ** 2}")

    @staticmethod
    def from_dict(d: dict) -> "YOLODetector":
        return YOLODetector(
            model=d.get("model", "yolov5n6"),
            classes=d.get("classes"),
            max_size=d.get("max_size", 1024),
            min_score=d.get("min_score", 0.3),
            filter_in_frame=d.get("filter_in_frame", False),
            min_area=d.get("min_area"),
        )

    def detect(self, image):
        h, w = image.shape[:2]
        if max(h, w) > self.max_size:
            scale = max(h, w) / self.max_size  # shape (1080, 1920), max_size=960 -> scale=1920/960 = 2
            dst_size = (int(w // scale), int(h // scale))
            image = cv2.resize(image, dst_size, interpolation=cv2.INTER_LINEAR)
        else:
            scale = 1

        # Run detection
        res = self.model(np.transpose(image, [2, 0, 1]))

        # Convert detections
        det = res.xyxy[0].cpu().numpy()
        rects, scores, labels = np.split(det, [4, 5], axis=1)
        labels = labels.ravel().astype(np.int32).tolist()
        scores = scores.ravel().tolist()

        # Convert coords to shapely Polygon
        geometries = map(lambda x: box(*x), rects * scale)

        # Generator of object instances
        all_detections = (
            ObjectObservation(
                geometry=geometry,
                score=score,
                label=label,
            )
            for geometry, label, score in zip(geometries, labels, scores)
        )

        # Filter objects that are in the frame
        if self.filter_in_frame:
            is_in_frame = lambda d: d.is_in_frame((h, w), margin=10)
            all_detections = filter(is_in_frame, all_detections)

        # Filter objects with sufficient size
        if self.min_area is not None and self.min_area > 0:
            sufficient_size = lambda d: d.geometry.area > self.min_area
            all_detections = filter(sufficient_size, all_detections)

        return list(all_detections)



"""
    SORT: A Simple, Online and Realtime Tracker
    Copyright (C) 2016-2020 Alex Bewley alex@bewley.ai

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# import os
import numpy as np
# import matplotlib
# import matplotlib.pyplot as plt
# import matplotlib.patches as patches
# from skimage import io

# import glob
# import time
# import argparse
from filterpy.kalman import KalmanFilter


def linear_assignment(cost_matrix):
    try:
        import lap
        _, x, y = lap.lapjv(cost_matrix, extend_cost=True)
        return np.array([[y[i], i] for i in x if i >= 0])  #
    except ImportError:
        from scipy.optimize import linear_sum_assignment
        x, y = linear_sum_assignment(cost_matrix)
        return np.array(list(zip(x, y)))


def iou_batch(bb_test, bb_gt):
    """
    From SORT: Computes IOU between two bboxes in the form [x1,y1,x2,y2]
    """
    bb_gt = np.expand_dims(bb_gt, 0)
    bb_test = np.expand_dims(bb_test, 1)

    xx1 = np.maximum(bb_test[..., 0], bb_gt[..., 0])
    yy1 = np.maximum(bb_test[..., 1], bb_gt[..., 1])
    xx2 = np.minimum(bb_test[..., 2], bb_gt[..., 2])
    yy2 = np.minimum(bb_test[..., 3], bb_gt[..., 3])
    w = np.maximum(0., xx2 - xx1)
    h = np.maximum(0., yy2 - yy1)
    wh = w * h
    o = wh / ((bb_test[..., 2] - bb_test[..., 0]) * (bb_test[..., 3] - bb_test[..., 1])
              + (bb_gt[..., 2] - bb_gt[..., 0]) * (bb_gt[..., 3] - bb_gt[..., 1]) - wh)
    return o


def convert_bbox_to_z(bbox):
    """
    Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form
      [x,y,s,r] where x,y is the centre of the box and s is the scale/area and r is
      the aspect ratio
    """
    w = bbox[2] - bbox[0]
    h = bbox[3] - bbox[1]
    x = bbox[0] + w / 2.
    y = bbox[1] + h / 2.
    s = w * h  # scale is just area
    r = w / float(h)
    return np.array([x, y, s, r]).reshape((4, 1))


def convert_x_to_bbox(x, score=None):
    """
    Takes a bounding box in the centre form [x,y,s,r] and returns it in the form
      [x1,y1,x2,y2] where x1,y1 is the top left and x2,y2 is the bottom right
    """
    w = np.sqrt(x[2] * x[3])
    h = x[2] / w
    if score is None:
        return np.array([x[0] - w / 2., x[1] - h / 2., x[0] + w / 2., x[1] + h / 2.]).reshape((1, 4))
    else:
        return np.array([x[0] - w / 2., x[1] - h / 2., x[0] + w / 2., x[1] + h / 2., score]).reshape((1, 5))


class KalmanBoxTracker(object):
    """
    This class represents the internal state of individual tracked objects observed as bbox.
    """
    count = 0

    def __init__(self, bbox, dt: float):
        """
        Initialises a tracker using initial bounding box.
        """
        # define constant velocity model
        self.kf = KalmanFilter(dim_x=7, dim_z=4)
        self.kf.F = np.array(
            [[1, 0, 0, 0, dt, 0, 0], [0, 1, 0, 0, 0, dt, 0], [0, 0, 1, 0, 0, 0, dt], [0, 0, 0, 1, 0, 0, 0],
             [0, 0, 0, 0, 1, 0, 0], [0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 1]]
        )
        self.kf.H = np.array(
            [[1, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0], [0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0]]
        )

        # 4x4 xysr
        self.kf.R = np.diag([1, 1, 5, 5]) * 0.1
        # 7x7 x,y,s,r,dx,dy,ds
        self.kf.P[4:, 4:] *= 100.  # give high uncertainty to the unobservable initial velocities
        self.kf.P *= 1
        # 7x7 noise
        self.kf.Q[-1, -1] *= 0.01
        self.kf.Q[4:, 4:] *= 0.01

        self.kf.x[:4] = convert_bbox_to_z(bbox)
        self.time_since_update = 0
        self.id = KalmanBoxTracker.count
        KalmanBoxTracker.count += 1
        self.history = []
        self.hits = 0
        self.hit_streak = 0
        self.age = 0

    def update(self, bbox):
        """
        Updates the state vector with observed bbox.
        """
        self.time_since_update = 0
        self.history = []
        self.hits += 1
        self.hit_streak += 1
        self.kf.update(convert_bbox_to_z(bbox))

    def predict(self):
        """
        Advances the state vector and returns the predicted bounding box estimate.
        """
        if (self.kf.x[6] + self.kf.x[2]) <= 0:
            self.kf.x[6] *= 0.0
        self.kf.predict()
        self.age += 1
        if self.time_since_update > 0:
            self.hit_streak = 0
        self.time_since_update += 1
        self.history.append(convert_x_to_bbox(self.kf.x))
        return self.history[-1]

    def get_state(self):
        """
        Returns the current bounding box estimate.
        """
        return convert_x_to_bbox(self.kf.x)


def associate_detections_to_trackers(detections, trackers, iou_threshold=0.3):
    """
    Assigns detections to tracked object (both represented as bounding boxes)

    Returns 3 lists of matches, unmatched_detections and unmatched_trackers
    """
    if len(trackers) == 0:
        return np.empty((0, 2), dtype=int), np.arange(len(detections)), np.empty((0, 5), dtype=int)

    iou_matrix = iou_batch(detections, trackers)

    if min(iou_matrix.shape) > 0:
        a = (iou_matrix > iou_threshold).astype(np.int32)
        if a.sum(1).max() == 1 and a.sum(0).max() == 1:
            matched_indices = np.stack(np.where(a), axis=1)
        else:
            matched_indices = linear_assignment(-iou_matrix)
    else:
        matched_indices = np.empty(shape=(0, 2))

    unmatched_detections = []
    for d, det in enumerate(detections):
        if d not in matched_indices[:, 0]:
            unmatched_detections.append(d)
    unmatched_trackers = []
    for t, trk in enumerate(trackers):
        if t not in matched_indices[:, 1]:
            unmatched_trackers.append(t)

    # filter out matched with low IOU
    matches = []
    for m in matched_indices:
        if iou_matrix[m[0], m[1]] < iou_threshold:
            unmatched_detections.append(m[0])
            unmatched_trackers.append(m[1])
        else:
            matches.append(m.reshape(1, 2))
    if len(matches) == 0:
        matches = np.empty((0, 2), dtype=int)
    else:
        matches = np.concatenate(matches, axis=0)

    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)


class Sort:
    def __init__(self, max_age: int = 1, min_hits: int = 3, iou_threshold: float = 0.3, dt: float = 1):
        """
        Sets key parameters for SORT
        """
        self.max_age = max_age
        self.min_hits = min_hits
        self.iou_threshold = iou_threshold
        self.trackers = []
        self.frame_count = 0
        self.dt = dt

    @staticmethod
    def from_dict(d: dict):
        return Sort(
            max_age=d.get("max_age", 1),
            min_hits=d.get("min_hits", 3),
            iou_threshold=d.get("iou", 0.3),
        )

    def update(self, detections=np.empty((0, 5))):
        """
        Params:
          detections - a numpy array of detections in the format [[x1,y1,x2,y2,score],[x1,y1,x2,y2,score],...]
        Requires: this method must be called once for each frame even with empty detections (use np.empty((0, 5)) for frames without detections).
        Returns a similar array, where the last column is the object ID.

        NOTE: The number of objects returned may differ from the number of detections provided.
        """
        self.frame_count += 1
        # get predicted locations from existing trackers.
        trackers = np.zeros((len(self.trackers), 5))
        to_del = []
        ret = []
        for t, trk in enumerate(trackers):
            pos = self.trackers[t].predict()[0]
            trk[:] = [pos[0], pos[1], pos[2], pos[3], 0]
            if np.any(np.isnan(pos)):
                to_del.append(t)
        trackers = np.ma.compress_rows(np.ma.masked_invalid(trackers))
        for t in reversed(to_del):
            self.trackers.pop(t)
        matched, unmatched_detections, unmatched_trackers = associate_detections_to_trackers(
            detections, trackers, self.iou_threshold
        )

        # update matched trackers with assigned detections
        for m in matched:
            self.trackers[m[1]].update(detections[m[0], :])

        # create and initialise new trackers for unmatched detections
        for i in unmatched_detections:
            trk = KalmanBoxTracker(detections[i, :], dt=self.dt)
            self.trackers.append(trk)
        i = len(self.trackers)
        for trk in reversed(self.trackers):
            d = trk.get_state()[0]
            if (trk.time_since_update < 1) and (trk.hit_streak >= self.min_hits or self.frame_count <= self.min_hits):
                ret.append(np.concatenate((d, [trk.id])).reshape(1, -1))  # +1 as MOT benchmark requires positive
            i -= 1
            # remove dead trackers
            if trk.time_since_update > self.max_age:
                self.trackers.pop(i)
        if len(ret) > 0:
            return np.concatenate(ret)
        return np.empty((0, 5))

# def parse_args():
#     """Parse input arguments."""
#     parser = argparse.ArgumentParser(description='SORT demo')
#     parser.add_argument('--display', dest='display', help='Display online tracker output (slow) [False]',action='store_true')
#     parser.add_argument("--seq_path", help="Path to detections.", type=str, default='data')
#     parser.add_argument("--phase", help="Subdirectory in seq_path.", type=str, default='train')
#     parser.add_argument("--max_age", 
#                         help="Maximum number of frames to keep alive a track without associated detections.", 
#                         type=int, default=1)
#     parser.add_argument("--min_hits", 
#                         help="Minimum number of associated detections before track is initialised.", 
#                         type=int, default=3)
#     parser.add_argument("--iou_threshold", help="Minimum IOU for match.", type=float, default=0.3)
#     args = parser.parse_args()
#     return args

# if __name__ == '__main__':
#   # all train
#   args = parse_args()
#   display = args.display
#   phase = args.phase
#   total_time = 0.0
#   total_frames = 0
#   colours = np.random.rand(32, 3) #used only for display
#   if(display):
#     if not os.path.exists('mot_benchmark'):
#       print('\n\tERROR: mot_benchmark link not found!\n\n    Create a symbolic link to the MOT benchmark\n    (https://motchallenge.net/data/2D_MOT_2015/#download). E.g.:\n\n    $ ln -s /path/to/MOT2015_challenge/2DMOT2015 mot_benchmark\n\n')
#       exit()
#     plt.ion()
#     fig = plt.figure()
#     ax1 = fig.add_subplot(111, aspect='equal')

#   if not os.path.exists('output'):
#     os.makedirs('output')
#   pattern = os.path.join(args.seq_path, phase, '*', 'det', 'det.txt')
#   for seq_detections_fn in glob.glob(pattern):
#     mot_tracker = Sort(max_age=args.max_age, 
#                        min_hits=args.min_hits,
#                        iou_threshold=args.iou_threshold) #create instance of the SORT tracker
#     seq_detections = np.loadtxt(seq_detections_fn, delimiter=',')
#     seq = seq_detections_fn[pattern.find('*'):].split(os.path.sep)[0]

#     with open(os.path.join('output', '%s.txt'%(seq)),'w') as out_file:
#       print("Processing %s."%(seq))
#       for frame in range(int(seq_detections[:,0].max())):
#         frame += 1 #detection and frame numbers begin at 1
#         detections = seq_detections[seq_detections[:, 0]==frame, 2:7]
#         detections[:, 2:4] += detections[:, 0:2] #convert to [x1,y1,w,h] to [x1,y1,x2,y2]
#         total_frames += 1

#         if(display):
#           fn = os.path.join('mot_benchmark', phase, seq, 'img1', '%06d.jpg'%(frame))
#           im =io.imread(fn)
#           ax1.imshow(im)
#           plt.title(seq + ' Tracked Targets')

#         start_time = time.time()
#         trackers = mot_tracker.update(detections)
#         cycle_time = time.time() - start_time
#         total_time += cycle_time

#         for d in trackers:
#           print('%d,%d,%.2f,%.2f,%.2f,%.2f,1,-1,-1,-1'%(frame,d[4],d[0],d[1],d[2]-d[0],d[3]-d[1]),file=out_file)
#           if(display):
#             d = d.astype(np.int32)
#             ax1.add_patch(patches.Rectangle((d[0],d[1]),d[2]-d[0],d[3]-d[1],fill=False,lw=3,ec=colours[d[4]%32,:]))

#         if(display):
#           fig.canvas.flush_events()
#           plt.draw()
#           ax1.cla()

#   print("Total Tracking took: %.3f seconds for %d frames or %.1f FPS" % (total_time, total_frames, total_frames / total_time))

#   if(display):
#     print("Note: to get real runtime results run without the option: --display")



import sys
from pathlib import Path
from typing import Union

import cv2
import rclpy
from cv_bridge import CvBridge
from rclpy.node import Publisher
from sensor_msgs.msg import Image
from std_msgs.msg import String

from fcw.client_python.client_common import CollisionWarningClient

collision_warning_client: Union[CollisionWarningClient, None] = None
publisher: Union[Publisher, None] = None

bridge = CvBridge()

# Configuration of the algorithm
config = Path("../../config/config.yaml")
# Camera settings - specific for the particular input
camera_config = Path("../../videos/video3.yaml")


def results_callback(data: dict):
    global publisher
    msg = String()
    print(data)
    msg.data = str(data)
    publisher.publish(msg)


def send_image_callback(image: Image):
    # convert received image to opencv
    cv_image = bridge.imgmsg_to_cv2(image, desired_encoding='passthrough')
    # color correction
    image_bgr = cv2.cvtColor(cv_image, cv2.COLOR_RGB2BGR)
    global collision_warning_client
    # or use the HTTP transport to the /image endpoint
    collision_warning_client.send_image(image_bgr, image.header.stamp.nanosec)


def main(args=None) -> None:
    """Main function."""

    rclpy.init(args=args)

    node = rclpy.create_node('client_ros2')
    global collision_warning_client, publisher

    publisher = node.create_publisher(String, "/results", 10)
    collision_warning_client = CollisionWarningClient(
        config=config, camera_config=camera_config, fps=30, results_callback=results_callback
    )
    subscriber = node.create_subscription(Image, "/image", send_image_callback, 10)

    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    except BaseException:
        print('Exception in node:', file=sys.stderr)
        raise
    finally:
        node.destroy_node()
        rclpy.shutdown()
        pass


if __name__ == "__main__":
    main()



"""
Early Collision Warning system
"""
from argparse import ArgumentParser, FileType
import statistics
from datetime import datetime
import cv2
import yaml
import time
import csv
import logging
import sys
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger("FCW example")

from fcw.core.collision import get_reference_points, ForwardCollisionGuard
from fcw.core.detection import detections_to_numpy
from fcw.core.sort import Sort
from fcw.core.vizualization import *
from fcw.core.yolo_detector import YOLODetector

from fcw.core.rate_timer import RateTimer

# os.environ["OPENCV_FFMPEG_CAPTURE_OPTIONS"] = "rtsp_transport;udp"

start_timestamp = datetime.now().strftime("%Y-%d-%m_%H-%M-%S")


def parse_arguments():
    parser = ArgumentParser()

    parser.add_argument("-c", "--config", type=FileType("r"), required=True, help="Collision warning config")
    parser.add_argument("--camera", type=FileType("r"), required=True, help="Camera settings")
    parser.add_argument("-o", "--output", type=str, help="Output video")
    parser.add_argument("--viz", action="store_true")
    parser.add_argument("--out_csv_dir", type=str, help="Output CSV dir", default=".")
    parser.add_argument("-p", "--out_prefix", type=str, help="Prefix of output csv file with measurements", default="fcw_example_test_")
    parser.add_argument("-t", "--play_time", type=int, help="Video play time in seconds", default=20)
    parser.add_argument("--fps", type=int, help="Video FPS", default=None)
    parser.add_argument("source_video", type=str, help="Video stream (file or url)")

    return parser.parse_args()


def main(args=None):
    args = parse_arguments()
    logger.info("Starting Forward Collision Guard")

    logger.info("Loading configuration file {cfg}".format(cfg=args.config.name))
    config_dict = yaml.safe_load(args.config)

    # Open video
    logger.info("Opening video {vid}".format(vid=args.source_video))
    video = cv2.VideoCapture(args.source_video)
    if not video.isOpened():
        raise Exception("Cannot open video file")
    if not args.fps:
        fps = video.get(cv2.CAP_PROP_FPS)
    else:
        fps = args.fps
    width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))
    shape = height, width
    logger.info("Video {W}x{H}, {fps} FPS".format(W=width, H=height, fps=fps))

    # Init object detector
    detector = YOLODetector.from_dict(config_dict.get("detector", {}))

    # Init image tracker
    logger.info("Initializing image tracker")
    tracker = Sort.from_dict(config_dict.get("tracker", {}))
    tracker.dt = 1 / fps

    # Init collision guard
    logger.info("Initializing Forward warning")
    guard = ForwardCollisionGuard.from_dict(config_dict.get("fcw", {}))
    guard.dt = 1 / fps  # Finish setup if the guard

    # Load camera calibration
    logger.info("Loading camera configuration {cfg}".format(cfg=args.camera.name))
    camera_dict = yaml.safe_load(args.camera)
    camera = Camera.from_dict(camera_dict)

    render_output = args.viz or args.output is not None
    if render_output:
        logger.warning("RENDERING OUTPUT - LOWER PERFOMANCE")

    if args.output is not None:
        output = cv2.VideoWriter(args.output, cv2.VideoWriter_fourcc(*"MP4V"), fps, camera.image_size)

    if args.viz:
        try:
            cv2.namedWindow("FCW")
        except Exception as ex:
            logger.debug(repr(ex))

    if render_output:
        # Prepare static stuff for vizualization
        logo = cog_logo((64, 64))
        coord_sys = draw_world_coordinate_system(camera.rectified_size, camera)
        coord_sys.putalpha(64)
        danger_zone = draw_danger_zone(camera.rectified_size, camera, guard.danger_zone)
        horizon = draw_horizon(camera.rectified_size, camera, width=1, fill=(255, 255, 0, 64))
        marker, marker_anchor = vehicle_marker_image(scale=3)

    delays = []
    timestamps = [
        ["start_timestamp_ns",
         "recv_timestamp_ns",
         "send_timestamp_ns",
         "end_timestamp_ns"]
    ]

    rate_timer = RateTimer(rate=fps, iteration_miss_warning=True)

    # FCW Loop
    start_time = time.time_ns()
    while time.time_ns() - start_time < args.play_time * 1.0e+9:
        ret, img = video.read()
        if not ret or img is None:
            logger.info("Video ended")
            break

        time0 = time.time_ns()
        img_undistorted = camera.rectify_image(img)
        time1 = time.time_ns()
        time_elapsed_s = (time1 - time0) * 1.0e-9
        logger.info(f"rectify_image time: {time_elapsed_s:.3f}")

        time0 = time.time_ns()

        # Detect object in image
        detections = detector.detect(img_undistorted)
        # Get bounding boxes as numpy array
        detections = detections_to_numpy(detections)
        # Update state of image trackers
        tracker.update(detections)
        # Represent trackers as dict  tid -> KalmanBoxTracker
        tracked_objects = {
            t.id: t for t in tracker.trackers
            if t.hit_streak > tracker.min_hits and t.time_since_update < 1 and t.age > 3
        }
        # Get 3D locations of objects
        ref_pt = get_reference_points(tracked_objects, camera, is_rectified=True)
        # Update state of objects in world
        guard.update(ref_pt)
        # Get list of current offenses
        dangerous_objects = guard.dangerous_objects()

        time1 = time.time_ns()
        logger.info(f"Delay: {(time1 - time0) * 1.0e-9:.3f}s")
        delays.append((time1 - time0))
        timestamps.append(
            [
                time0,
                time0,
                time1,
                time1
            ]
        )

        if render_output:
            # Vizualization
            base_undistorted = Image.fromarray(img_undistorted[..., ::-1], "RGB").convert("L").convert("RGBA")
            # Base layer is the camera image
            base = Image.fromarray(img[..., ::-1], "RGB").convert("RGBA")
            # Layers showing various information
            sz = base_undistorted.size
            layers = [
                (coord_sys, None),
                (danger_zone, None),
                (horizon, None),
                (draw_image_trackers(sz, tracker.trackers), None),
                (draw_world_objects(sz, camera, guard.objects.values()), None),
            ]
            # Compose layers together
            compose_layers(base_undistorted, *layers)
            O = list(guard.label_objects(include_distant=True))
            w, h = base.size
            w1, h1 = base_undistorted.size
            compose_layers(
                base,  # Original image
                (tracking_info((w, 16), O), (0, 0)),
                (mark_vehicles(camera.image_size, guard.objects.values(), camera, marker, marker_anchor), None),
                (logo, (8, 16 + 8)),
                (base_undistorted, (8, h - h1 - 8)),  # Pic with rectified image and vizualized trackers
            )
            # Convert to OpenCV for display
            cv_image = np.array(base.convert("RGB"))[..., ::-1]

        if args.viz:
            try:
                # Display the image
                cv2.imshow("FCW", cv_image)
                cv2.waitKey(1)
            except Exception as ex:
                logger.debug(repr(ex))

        if args.output is not None:
            output.write(cv_image)

        rate_timer.sleep()  # sleep until next frame should be sent (with given fps)

    if args.output is not None:
        output.release()

    logger.info(f"-----")
    end_time = time.time_ns()
    logger.info(f"Total streaming time: {(end_time - start_time) * 1.0e-9:.3f}s")
    logger.info(f"Delay median: {statistics.median(delays) * 1.0e-9:.3f}s")

    if args.out_csv_dir is not None:
        out_csv_filename = f'{args.out_prefix}'
        out_csv_filepath = os.path.join(args.out_csv_dir, out_csv_filename + ".csv")
        with open(out_csv_filepath, "w", newline='') as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerows(timestamps)

    try:
        cv2.destroyAllWindows()
    except Exception as ex:
        logger.debug(repr(ex))


if __name__ == "__main__":
    main()



from setuptools import setup

package_name = 'client_ros2'

setup(
    name=package_name,
    version='0.0.0',
    packages=[package_name],
    data_files=[
        ('share/ament_index/resource_index/packages',
            ['resource/' + package_name]),
        ('share/' + package_name, ['package.xml']),
    ],
    install_requires=['setuptools'],
    zip_safe=True,
    maintainer='Petr Kleparnik',
    maintainer_email='ikleparnik@fit.vutbr.cz',
    description='CollisionWarningService client',
    license='TODO: License declaration',
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'client_node = client_ros2.client_node:main',

        ],
    },
)



"""
"""

from typing import Iterable, List

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from shapely.geometry import LineString, Polygon
import os.path
from os import path

from fcw.core import sort
from fcw.core.collision import PointWorldObject, ObjectStatus
from fcw.core.geometry import Camera

# Font for OSD
if path.exists("../../data/UbuntuMono-R.ttf"):
    _font = ImageFont.truetype("../../data/UbuntuMono-R.ttf", 14, encoding="unic")
elif path.exists("data/UbuntuMono-R.ttf"):
    _font = ImageFont.truetype("data/UbuntuMono-R.ttf", 14, encoding="unic")
elif path.exists("/usr/share/fonts/truetype/ubuntu/UbuntuMono-R.ttf"):
    _font = ImageFont.truetype("/usr/share/fonts/truetype/ubuntu/UbuntuMono-R.ttf", 14, encoding="unic")
else:
    _font = None

# def segmentize(p: LineString, max_dist=10):
#     pts = []
#     for a, b in windowed(p.coords, n=2):
#         seg = LineString([a, b])
#         f = np.linspace(0, seg.length, ceil(seg.length / max_dist), endpoint=False)
#         _pts = [seg.interpolate(x) for x in f]
#         pts.extend(_pts)
#     return LineString(pts)


def draw_horizon(size: tuple, cam: Camera, **kwargs):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)
    x = list(cam.horizon.coords)
    draw.line(x, **kwargs)
    return image


def compose_layers(base: Image.Image, *layers: Iterable[tuple]):
    for l, dest in layers:
        base.alpha_composite(l, dest or (0, 0))


def draw_image_trackers(
    size: tuple,
    trackers: List[sort.KalmanBoxTracker],
):
    """

    """
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    for t in trackers:
        x1, y1, x2, y2 = t.get_state()[0]
        color = (0, 255, 0, 64)
        outline = (0, 255, 0, 128)
        if t.age < 3 or t.hit_streak == 0:  # TODO: call it.is_reliable()
            color = (255, 255, 0, 32)
            outline = None
        draw.rectangle((x1, y1, x2, y2), fill=color, outline=outline)

    return image


def draw_world_objects(
    size: tuple,
    camera: Camera,
    objects: Iterable[PointWorldObject]
):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    # for o in objects:
    #     x1, y1, x2, y2 = tracked_objects[tid].get_state()[0]
    #     # objects_draw.rectangle([x1, y1, x2, y2], outline=(255, 0, 0), width=2)
    #     objects_draw.rectangle((x1, y1, x2, y2), fill=(255, 0, 0, 64))
    #     dist = Point(o.location).distance(guard.vehicle_zone)
    #     info = f"{dist:.1f} m"
    #     objects_draw.text(
    #         (0.5 * (x1 + x2), 0.5 * (y1 + y2)), info, align="center", font=font,
    #         stroke_fill=(255, 255, 255), stroke_width=1, fill=(0, 0, 0)
    #         )

    for o in objects:

        X = np.atleast_2d([o.kf.x[0, 0], o.kf.x[3, 0], 0])
        scr_loc, _ = camera.project_points(X)
        if scr_loc.size > 0:
            x, y = scr_loc[0]
            draw.line([(x - 10, y), (x + 10, y)], fill=(255, 255, 0, 128), width=3)
            draw.line([(x, y - 10), (x, y + 10)], fill=(255, 255, 0, 128), width=3)

        X = np.array(o.future_path(5).coords)
        n = X.shape[0]
        X = np.hstack([X, np.zeros((n, 1))])
        scr_loc, _ = camera.project_points(X, near=5)
        scr_loc = list(map(tuple, scr_loc))
        draw.line(scr_loc, fill=(0, 255, 0, 255), width=1)

    return image


def draw_danger_zone(size: tuple, camera: Camera, zone: Polygon):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    front = Polygon(
        [
            [1, -10],
            [1, 10],
            [50, 10],
            [50, -101],
        ]
    )

    X = np.array(zone.intersection(front).boundary.coords)
    n = X.shape[0]
    X = np.hstack([X, np.zeros((n, 1))])
    scr_loc, _ = camera.project_points(X, near=-100)
    scr_loc = list(map(tuple, scr_loc))
    draw.polygon(scr_loc, fill=(255, 255, 0, 32), outline=(255, 255, 0, 128))

    return image


def draw_world_coordinate_system(size: tuple, camera: Camera):
    image = Image.new("RGBA", size)
    draw = ImageDraw.Draw(image)

    def draw_line_string(ls: LineString, accuracy: float = 0.01, **kwargs):
        n = int(np.ceil(1 / accuracy))
        coords = [
            ls.interpolate(t, normalized=True).xy for t in np.linspace(0, 1, n)
        ]
        X = np.hstack(coords)
        X = np.vstack([X, np.zeros((1, n))])
        x, _ = camera.project_points(X.T, near=1)
        draw.line(list(map(tuple, x)), **kwargs)

    for x in np.linspace(-20, 20, 41):
        draw_line_string(LineString([(x, -20), (x, 20)]), fill=(255, 0, 0))

    for y in np.linspace(-20, 20, 41):
        draw_line_string(LineString([(-20, y), (20, y)]), fill=(0, 255, 0))

    return image


def tracking_info(
    size: tuple,
    object_status: List[ObjectStatus]
):
    image = Image.new("RGBA", size, color=(0, 0, 0, 255))
    draw = ImageDraw.Draw(image)

    info_text = f"Tracking {len(object_status)}"
    draw.text((8, 0), info_text, fill=(255, 255, 255), font=_font)

    caution_status = any(
        s.crosses_danger_zone for s in object_status
    )

    warning_status = any(
        s.is_in_danger_zone for s in object_status
    )

    danger_status = any(
        s.time_to_collision > 0 for s in object_status if s.time_to_collision is not None
    )

    caution_color = (255, 255, 0) if caution_status else (64, 64, 64)
    draw.text((160, 0), "CAUTION", fill=caution_color, font=_font, align="left")

    warning_color = (255, 255, 0) if warning_status else (64, 64, 64)
    draw.text((260, 0), "WARNING", fill=warning_color, font=_font, align="left")

    danger_color = (255, 0, 0) if danger_status else (64, 64, 64)
    draw.text((360, 0), "DANGER", fill=danger_color, font=_font, align="left")

    if danger_status:
        ttc = min(
            s.time_to_collision for s in object_status if s.time_to_collision is not None
        )
        draw.text((460, 0), f"ttc = {ttc:0.1f} s", fill=(255, 0, 0), font=_font, align="left")

    return image


def cog_logo(size: tuple = (256, 256)):
    """
    Cognitechna logo image
    """
    logo = Image.open("../../data/cog_logo.png").convert(
        "RGBA"
    )  # FIXME location data in the package not relative to `pwd`
    w, h = logo.size
    cx, cy = w / 2, h / 2
    sz = 155
    box = cx - sz, cy - sz, cx + sz, cy + sz
    logo = logo.resize(size, box=box, reducing_gap=True, resample=Image.LANCZOS)

    bg = Image.new("RGBA", size, (255, 255, 255))
    bg.alpha_composite(logo)

    drw = ImageDraw.ImageDraw(bg)
    drw.rectangle((0, 0, size[0] - 1, size[1] - 1), fill=None, outline=(0, 0, 0, 255), width=1)

    return bg


def vehicle_marker_image(scale: int = 1):
    marker_image = Image.open("../../data/marker.png")
    w, h = marker_image.size
    return marker_image.resize((w * scale, h * scale), Image.NEAREST), (7 * scale, 0)


def mark_vehicles(
    size: tuple, objects: Iterable[PointWorldObject], camera: Camera, marker: Image, anchor: tuple = (0, 0)
):
    image = Image.new("RGBA", size, color=(0, 0, 0, 0))
    ax, ay = anchor
    # loc (N,2) xy
    for o in objects:
        X = np.atleast_2d([o.kf.x[0, 0], o.kf.x[3, 0], 0])
        scr_loc, _ = camera.project_points(X, near=1, to_rectified=False)
        if scr_loc.shape[0] > 0:
            x, y = scr_loc[0]
            image.paste(marker, (int(x - ax), int(y - ay)))
    return image



from typing import Dict
from dataclasses import dataclass
import numpy as np
from filterpy.common import Q_discrete_white_noise
from filterpy.kalman import KalmanFilter
from shapely.geometry import LineString, Point, Polygon, box
import logging
logger = logging.getLogger(__name__)

from fcw.core.geometry import *


def F_matrix(dt):
    dt2 = 0.5 * dt ** 2
    return np.array(
        [[1, dt, dt2, 0, 0, 0],
         [0, 1, dt, 0, 0, 0],
         [0, 0, 1, 0, 0, 0],
         [0, 0, 0, 1, dt, dt2],
         [0, 0, 0, 0, 1, dt],
         [0, 0, 0, 0, 0, 1]], dtype=np.float32
    )


def object_tracker(x_init, dt: float = 1):
    kf = KalmanFilter(dim_x=6, dim_z=2)
    kf.F = F_matrix(dt)

    kf.H = np.array(
        [[1, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 0, 0]]
    )  # Measurement function

    kf.P = np.diag([1, 2, 400, 1, 2, 400]) * 10
    z_std = 2
    kf.R = np.diag([z_std ** 2, z_std ** 2])  # 1 standard
    kf.Q = Q_discrete_white_noise(dim=3, dt=dt, var=0.5e-1 ** 2, block_size=2)  # process uncertainty
    kf._alpha_sq = 1
    x, y = x_init
    kf.x[0] = x
    kf.x[3] = y
    return kf


def covariance(xy: np.ndarray, sigma: float = 0.1, scale: float = 0.1):
    d = np.linalg.norm(xy)
    x, y = xy
    th = np.arctan2(y, x)
    c, s = np.cos(th), np.sin(th)
    M = np.array([[c, -s], [s, c]])
    return M @ (np.diag([1, scale]) * sigma * d)


class PointWorldObject:
    """
    Simplest abstraction of world objects - just a location

    One can implement 
    """

    def __init__(self, xyz: np.ndarray, dt: float):
        self.kf = object_tracker(xyz[:2], dt=dt)
        self.xy = None
        self.vxvy = None

    def update(self, location=None):
        self.kf.predict()
        self.kf.update(location, R=covariance(location, sigma=0.1, scale=0.1))
        self.xy = np.dot(self.kf.H, self.kf.x).T[0]
        self.vxvy = np.dot(np.array([[0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 1, 0]]), self.kf.x).T[0]

    @property
    def location(self):
        return self.xy

    @property
    def distance(self):
        if self.xy is None:
            return np.inf
        else:
            return np.linalg.norm(self.xy)

    @property
    def relative_speed(self):
        if self.vxvy is None: return 0
        return np.linalg.norm(self.vxvy)

    def future_path(self, length: float = 1, dt: float = 0.1):
        x = self.kf.x
        F = F_matrix(dt)
        t = 0
        X = [x]
        while t < length:
            x = np.dot(F, x)
            t += dt
            X.append(x)
        xy = self.kf.H @ np.hstack(X)  # (2, N)
        return LineString(xy.T)


def get_reference_points(trackers: dict, camera: Camera, *, is_rectified: bool):
    """
    Convert 2D observation to 3D
    """
    if not trackers: return dict()

    # (xyxy) -> (rx,ry)
    R = np.array(
        [[0.5, 0, 0.5, 0],
         [0, 0, 0, 1]],
    )
    # image space bounding boxes of objects
    bb = np.vstack([tracker.get_state() for tracker in trackers.values()]).T
    img_rp = R @ bb  # (2,N) 2D ref points in distorted image

    if not is_rectified:
        # If trackers are used on non-rectified image
        img_rp = camera.rectify_points(img_rp.T).T

    # points are in cam.K_new camera - normalize and convert to (3,N)
    n = img_rp.shape[1]
    img_rp_h = np.vstack([img_rp, np.ones((1, n))])
    norm_rp = inv(camera.K_new) @ img_rp_h  # (3,N) normalized xyz in camera space

    norm_rp = np.vstack([norm_rp, np.ones((1, n))])  # (4,N) homogeneous in 3D

    X = camera.RT_inv @ norm_rp  # (3,N)
    O = camera.RT_inv[:, -1].reshape(3, -1)
    S = X - O
    plane_normal = np.atleast_2d([0, 0, 1])
    t = (plane_normal @ O[:3]) / (plane_normal @ S[:3])
    world_rp = O - t * S

    return dict(zip(trackers.keys(), world_rp.T))  # tid -> (x,y,z)


class ForwardCollisionGuard:
    def __init__(
        self,
        danger_zone: Polygon,
        vehicle_zone: Polygon,
        safety_radius: float = 25,
        prediction_length: float = 1,
        prediction_step: float = 0.1,
        dt: float = 1,
    ):
        self.dt = dt
        self.objects: Dict[int, PointWorldObject] = dict()
        self.danger_zone = danger_zone
        self.vehicle_zone = vehicle_zone
        self.safety_radius = safety_radius  # m
        self.prediction_length = prediction_length
        self.prediction_step = prediction_step

    @staticmethod
    def from_dict(d):
        zone = Polygon(d.get("danger_zone"))
        length, width = d.get("vehicle_length", 4), d.get("vehicle_width", 1.8)
        vehicle_zone = box(-length / 2, -width / 2, length / 2, width / 2).buffer(0.5, resolution=4)

        return ForwardCollisionGuard(
            danger_zone=zone,
            vehicle_zone=vehicle_zone,
            safety_radius=d.get("safety_radius", 30),
            prediction_length=d.get("prediction_length", 1),
            prediction_step=d.get("prediction_step", 0.1),
        )

    def update(self, ref_points: dict):
        """
        Update state of objects tracked in world space
        """
        # Sync world trackers with image trackers
        for tid in list(self.objects.keys()):
            if tid not in ref_points:
                self.objects.pop(tid)
                logger.info(f"Tracking object with id {tid} lost")

        for tid in ref_points.keys():
            if tid not in self.objects:
                logger.info("Tracking object with id {tid}".format(tid=tid))
                self.objects[tid] = PointWorldObject(ref_points[tid], self.dt)
            else:
                self.objects[tid].update(ref_points[tid][:2])

    def dangerous_objects(self):
        """
        Check future paths of objects and filter dangerous ones
        """
        return {
            tid: obj for tid, obj in self.objects.items()
            if obj.distance < self.safety_radius and obj.future_path(
                self.prediction_length, self.prediction_step
            ).intersects(self.danger_zone)
        }

    def label_objects(
        self,
        include_distant: bool = False,
    ):
        """
        Check future paths of objects and filter dangerous ones
        """
        for tid, obj in self.objects.items():
            if obj.xy is None: continue

            loc = Point(obj.location)
            dist = loc.distance(self.vehicle_zone)

            if dist > self.safety_radius and not include_distant:
                continue

            path = obj.future_path(self.prediction_length, self.prediction_step)
            collision_point_distance = intersection_point(path, self.vehicle_zone.boundary)
            if collision_point_distance is not None:
                ttc = (collision_point_distance / path.length) * self.prediction_length
            else:
                ttc = None

            yield ObjectStatus(
                distance=dist,
                location=loc,
                path=path,
                is_in_danger_zone=self.danger_zone.contains(loc),
                crosses_danger_zone=path.crosses(self.danger_zone),
                time_to_collision=ttc,
            )


@dataclass
class ObjectStatus:
    # Distance from the reference point to vehicle zone
    distance: float
    # Location relative to vehicle reference point
    location: Point
    path: LineString
    # Flag indicating object in danger zone
    is_in_danger_zone: bool
    # Flag indicating if object path crosses danger zone
    crosses_danger_zone: bool
    # Time for reference point to reach vehicle zone. If None, does not reach vehicle
    time_to_collision: float

    @property
    def is_colliding(self):
        return self.time_to_collision is not None

    @property
    def is_dangerous(self):
        """
        Dangerous objects
        """
        danger = self.is_in_danger_zone
        collision = self.time_to_collision is not None and self.time_to_collision < 1
        return danger or collision


from more_itertools import pairwise


def intersection_point(ls: LineString, p: LineString):
    coords = ls.coords
    d = 0
    for a, b in pairwise(coords):
        l = LineString([a, b])
        pt = l.intersection(p)
        if not pt.is_empty:
            return Point(a).distance(pt) + d
        d += l.length



from multiprocessing import Queue
from threading import Thread, Event
from queue import Empty
import time
import logging

from era_5g_interface.interface_helpers import LatencyMeasurements
from fcw.core.collision import *
from fcw.core.detection import *
from fcw.core.sort import Sort
from fcw.core.yolo_detector import YOLODetector

logger = logging.getLogger(__name__)


class CollisionWorker(Thread):
    def __init__(
        self,
        image_queue: Queue,
        sio,
        config: dict,
        camera_config: dict,
        fps: float,
        **kw
    ):
        super().__init__(**kw)
        self.stop_event = Event()
        self.image_queue = image_queue
        self.sio = sio
        self.frame_id = 0
        self.latency_measurements = LatencyMeasurements()

        logger.info("Initializing object detector")
        self.detector = YOLODetector.from_dict(config.get("detector", {}))
        logger.info("Initializing image tracker")
        self.tracker = Sort.from_dict(config.get("tracker", {}))
        logger.info("Initializing forward collision guard")
        self.guard = ForwardCollisionGuard.from_dict(config.get("fcw", {}))
        self.guard.dt = 1 / fps
        logger.info("Initializing camera calibration")
        self.camera = Camera.from_dict(camera_config)

    def stop(self):
        self.stop_event.set()

    def __del__(self):
        logger.info("Delete object detector")
        del self.detector

    def run(self):
        """
        Periodically reads images from python internal queue process them.
        """

        logger.info(f"{self.name} thread is running.")

        while not self.stop_event.is_set():
            # Get image and metadata from input queue
            try:
                metadata, image = self.image_queue.get(block=True, timeout=1)
            except Empty:
                continue
            metadata["timestamp_before_process"] = time.perf_counter_ns()
            self.frame_id += 1
            # logger.info(f"Worker received frame id: {self.frame_id} {metadata['timestamp']}")
            try:
                detections = self.process_image(image)
                metadata["timestamp_after_process"] = time.perf_counter_ns()
                self.publish_results(detections, metadata)
            except Exception as e:
                logger.error(f"Exception with image processing: {repr(e)}")

        logger.info(f"{self.name} thread is stopping.")

    def process_image(self, image):
        # Detect object in image
        detections = self.detector.detect(image)
        # Get bounding boxes as numpy array
        detections = detections_to_numpy(detections)
        # Update state of image trackers
        self.tracker.update(detections)
        # Represent trackers as dict  tid -> KalmanBoxTracker
        tracked_objects = {
            t.id: t for t in self.tracker.trackers
            if t.hit_streak > self.tracker.min_hits and t.time_since_update < 1
        }
        # Get 3D locations of objects
        ref_points = get_reference_points(tracked_objects, self.camera, is_rectified=True)
        # Update state of objects in world
        self.guard.update(ref_points)

        return tracked_objects

    def publish_results(self, tracked_objects, metadata):
        """
        Publishes the results to the robot

        Args:
            tracked_objects (_type_): The results of the detection.
            metadata (_type_): NetApp-specific metadata related to processed image.
        """
        # Get list of current offenses
        dangerous_objects = self.guard.dangerous_objects()
        detections = dict()
        if tracked_objects is not None:
            for tid, t in tracked_objects.items():
                x1, y1, x2, y2 = t.get_state()[0]
                det = dict()
                det["bbox"] = [x1, y1, x2, y2]
                det["dangerous_distance"] = 0

                if tid in dangerous_objects.keys():
                    dist = Point(dangerous_objects[tid].location).distance(self.guard.vehicle_zone)
                    det["dangerous_distance"] = dist
                detections[tid] = det

                # det["class"] = result.label
                # det["class_name"] = self.detector.model.names[result.label]

            # TODO:check timestamp exists
            result = {"timestamp": metadata["timestamp"],
                      "recv_timestamp": metadata["recv_timestamp"],
                      "timestamp_before_process": metadata["timestamp_before_process"],
                      "timestamp_after_process": metadata["timestamp_after_process"],
                      "send_timestamp": time.perf_counter_ns(),
                      "detections": detections}
            self.sio.emit('message', result, namespace='/results', to=metadata["websocket_id"])
